#include <KernelExport.h>
#include <OS.h>
#include <Drivers.h>
#include <Errors.h>
#include <USB.h>

#ifndef __CREATIVE_MP3_PLAYER__
#define __CREATIVE_MP3_PLAYER__
#include "PDE_DEFS.h"
#define CREATIVE_MP3_ID "CREATIVE MP3 PLAYER "
#define CREATIVE_ZEN_MICRO_ID  (uint16) 0x411e
#define CREATIVE_ZEN_TOUCH_ID  (uint16)0x411b
#define CREATIVE_ZEN_TOUCH2_ID  (uint16)0x411d

#define CREATIVE_ZEN_JUKEBOX2_ID  (uint16)0x4100
#define CREATIVE_ZEN_JUKEBOX3_ID  (uint16)0x4101
#define CREATIVE_ZEN_ZEN_ID  (uint16)0x4108
#define CREATIVE_ZEN_ZENUSB2_ID  (uint16)0x410b
#define CREATIVE_ZEN_NX_ID  (uint16)0x4109
#define CREATIVE_ZEN_EXTRA_ID  (uint16)0x4110
#define CREATIVE_ZEN_DJUKEBOX_ID  (uint16)0x4111
#define CREATIVE_ZEN_DJUKEBOX2_ID  (uint16)0x4126
#define CREATIVE_ZEN_PCKET_ID  (uint16)0x4127
#define CREATIVE_ZEN_SLK_ID  (uint16)0x4136

#define PROTO_PDE  (uint16) 0xFF
#define CREATIVE_MANUFACTURE_CODE (uint16) 0x041e
#define CREATIVE_SUBCLASS  (uint16) 0xFF
#define CREATIVE_SUBCLASS  (uint16) 0xFF

typedef struct creative_mp3player_buffer njb_packet;
struct creative_mp3player_buffer{
	njb_packet* next;
	void* buff;	
	size_t size;
	area_id packet_id;
};

typedef struct creative_mp3player_struct
{
	struct creative_mp3player_struct *next;
	int open;
	int number;
	
	const usb_device *dev;
	const usb_configuration_info *conf;
	/* lock and notify semaphores and scratchpad buffer */
	sem_id lock;
	size_t size;
	area_id buffer_id;
	uint32 num_packets;
	njb_packet *packet;
	njb_packet *tail;
	// used for logging outbound packets
	sem_id  lock_send;
	size_t  size_send;
	njb_packet *packet_send;
	njb_packet *packet_send_tail;
	//
	int nr_callbacks;
	/* callback return values */
	status_t njb_status;
	status_t status;
} mp3player_type;

typedef struct supported_device_list{
 	int i;
 	char *dev_name;
} type_supported_device;
 
typedef struct
{
	char signature[4];
	
	struct _usbdev *parent;
	int refcount;
	int active; /* device is live and well */
	
	sem_id lock;
	
	char *name;
} usb_device0;

//constants.c njb3 commands ...
extern const unsigned char njb3_ping0[];
extern unsigned char njb3_ping1[];
extern const unsigned char usb11_cmd_header[];
extern const unsigned char usb20_cmd_header[];
extern const unsigned char get_frame[];
extern const unsigned char get_disk_usage[];
extern char njb3_get_track_tags[];
extern const unsigned char njb3_read_keys[];
extern unsigned char njb3_request_file[];
extern const unsigned char njb3_create_file[];
extern unsigned char njb3_send_file_chunk[];
extern unsigned char njb3_send_file_complete[];
extern unsigned char njb3_delitem[];
// util.c
extern void from_32bit_to_njb3_bytes(uint32 val, unsigned char *dp);
extern void from_16bit_to_njb3_bytes(uint16 val, unsigned char *dp);
extern uint16 njb3_bytes_to_16bit(unsigned char *dp);
void print_error(char *tag,status_t st);
extern unsigned char swap(unsigned char p);
extern void output_log(char *buff);
// creative_mp3player.c
extern sem_id dev_list_lock;
extern usb_module_info *usb;
extern status_t send_usb_command(const usb_device *,uint8 );
extern mp3player_type *cur_dev;

//  commands.c
extern mp3player_type *device_list;
extern int bBurstDataTransfer;
extern status_t recv(void *cookie, void (*callback)(void *cookie, uint32 status, 
								  void *data, uint32 actual_len)
								  );
extern status_t send_njb_command(uint32 flags, void** cookie,
			unsigned char* command,size_t clength);
extern status_t send_njb_data(uint32 flags, void** cookie,
			unsigned char* command,size_t clength);
// copy data to userspace
extern void copy_to_userspace(char *buff,mp3player_type *ptr);

void callback_burst_recv(void *cookie, uint32 status, 
								  void *data, uint32 actual_len);
void callback_normal_recv(void *cookie, uint32 status, 
								  void *data, uint32 actual_len);
void send_njb_command_callback(void *cookie, uint32 status, 
								  void *data, uint32 actual_len);

#endif