/***
 *	Creative MP3 Player driver
 *  **********************************
 *	Author Atsushi Majima
 *  Copyright 2008 - 
 */
#include "PDE.h"

sem_id dev_list_lock = -1;
//static int device_count = 0;

static char **device_names = NULL;
static int16 iCounter_Device=0;
static const char *drivername = "creative_mp3player";
static const char *basename = "bus/usb";
usb_module_info *usb = NULL;

static status_t device_removed(void *cookie);
static status_t device_added(const usb_device *dev, void **cookie);

mp3player_type *device_list=NULL;

#ifndef PACKET_SIZE
#define PACKET_SIZE(x) (x->endpoint.max_packet_size)
#endif

#define TS	16	/* 4 ms buffers */
#define NB	8

static usb_notify_hooks notify_hooks = 
{
	&device_added,
	&device_removed
};

/**
 * Device Descriptors
 * Many PDE 1.0 devices are supported ...(due to the LIBNJB)
 */
usb_support_descriptor supported_devices[] =
{
	{ CREATIVE_SUBCLASS, CREATIVE_SUBCLASS,PROTO_PDE, CREATIVE_MANUFACTURE_CODE , CREATIVE_ZEN_MICRO_ID},
	{ CREATIVE_SUBCLASS, CREATIVE_SUBCLASS,PROTO_PDE, CREATIVE_MANUFACTURE_CODE , CREATIVE_ZEN_TOUCH_ID },
	{ CREATIVE_SUBCLASS, CREATIVE_SUBCLASS,PROTO_PDE, CREATIVE_MANUFACTURE_CODE , CREATIVE_ZEN_TOUCH2_ID },
	{ CREATIVE_SUBCLASS, CREATIVE_SUBCLASS,PROTO_PDE, CREATIVE_MANUFACTURE_CODE , CREATIVE_ZEN_JUKEBOX2_ID },
	{ CREATIVE_SUBCLASS, CREATIVE_SUBCLASS,PROTO_PDE, CREATIVE_MANUFACTURE_CODE , CREATIVE_ZEN_JUKEBOX3_ID },
	{ CREATIVE_SUBCLASS, CREATIVE_SUBCLASS,PROTO_PDE, CREATIVE_MANUFACTURE_CODE , CREATIVE_ZEN_ZEN_ID },
	{ CREATIVE_SUBCLASS, CREATIVE_SUBCLASS,PROTO_PDE, CREATIVE_MANUFACTURE_CODE , CREATIVE_ZEN_ZENUSB2_ID },
	{ CREATIVE_SUBCLASS, CREATIVE_SUBCLASS,PROTO_PDE, CREATIVE_MANUFACTURE_CODE , CREATIVE_ZEN_NX_ID },
	{ CREATIVE_SUBCLASS, CREATIVE_SUBCLASS,PROTO_PDE, CREATIVE_MANUFACTURE_CODE , CREATIVE_ZEN_EXTRA_ID },
	{ CREATIVE_SUBCLASS, CREATIVE_SUBCLASS,PROTO_PDE, CREATIVE_MANUFACTURE_CODE , CREATIVE_ZEN_DJUKEBOX_ID },
	{ CREATIVE_SUBCLASS, CREATIVE_SUBCLASS,PROTO_PDE, CREATIVE_MANUFACTURE_CODE , CREATIVE_ZEN_DJUKEBOX2_ID },
	{ CREATIVE_SUBCLASS, CREATIVE_SUBCLASS,PROTO_PDE, CREATIVE_MANUFACTURE_CODE , CREATIVE_ZEN_PCKET_ID },
	{ CREATIVE_SUBCLASS, CREATIVE_SUBCLASS,PROTO_PDE, CREATIVE_MANUFACTURE_CODE , CREATIVE_ZEN_SLK_ID },
};

/**
 * relative path to the supported devices  
 */
static const char *my_device_name[]={
	"bus/usb/2/1",
	NULL
} ;

//static const char **my_device_name = NULL;
								  
/**
 * init_hardware - called once the first time the driver is loaded
 */
_EXPORT status_t
init_hardware (void)
{
	output_log("init_hardware called\n");
	return B_OK;
}


/**
 * init_driver - optional function - called every time the driver is loaded.
 * load up USB bus manager module and register the creative driver
 */
_EXPORT status_t init_driver (void)
{
	uint8 i;
	char buff[1024];
	//kernel_debugger("init_driver");

	dprintf(CREATIVE_MP3_ID "init_driver(), built %s %s\n", __DATE__, __TIME__);
	
	output_log("init driver was called\n");
	
	if(load_driver_symbols(drivername) != B_OK) {
		output_log( CREATIVE_MP3_ID "no symbols for you!\n");
		return B_ERROR;
	}
		
	if(get_module(B_USB_MODULE_NAME,(module_info**) &usb) != B_OK){
		dprintf(CREATIVE_MP3_ID "cannot get module \"%s\"\n",B_USB_MODULE_NAME);
		output_log("after get_module");
		return B_ERROR;
	} 
	
	if((dev_list_lock = create_sem(1,"dev_list_lock")) < 0){
		put_module(B_USB_MODULE_NAME);
		kprintf("after put_module");
		return dev_list_lock;
	}
	
	
	if( usb->register_driver(drivername,supported_devices, 
		sizeof(supported_devices)/sizeof(usb_support_descriptor) , NULL) != B_OK ){		
		output_log("ouch , driver couldn't be registered !!!");
		return B_ERROR;
	}
	if( usb->install_notify(drivername, &notify_hooks) != B_OK){
		output_log("ouch , driver couldn't be notified !!!");
		return B_ERROR;
	}
	output_log("init driver end\n");
	return B_OK;
}

/**
 *
 */
static void free_all_packets(njb_packet **packet){
	njb_packet *ptr=*packet;
	while(ptr!=NULL){
		if(ptr->packet_id != 0)
			delete_area(ptr->packet_id);
		else
			free(ptr->buff);
		ptr=ptr->next;
	}
}

/**
 * uninit_driver - optional function - called every time the driver
 * is unloaded
 */
_EXPORT void uninit_driver (void)
{
	mp3player_type *ptr=device_list;
	
	output_log("uninit_driver was called\n");

/*
	acquire_sem(dev_list_lock);	
	while(ptr != NULL){
		usb->set_configuration(ptr->dev,NULL);
		free_all_packets(&(ptr->packet));
		free(ptr);
	}
	device_list=NULL;
*/
	output_log("uninit_driver end ...\n");	
}

	
/***
 * my_device_open - handle open() calls
 */
static status_t
zen_device_open (const char *name, uint32 flags, void** cookie)
{
	char buff[1024];
	mp3player_type *ptr;
	mp3player_type *ptr1;
	sprintf(buff,"zen_device_open ...%s\n",name);
	output_log(buff);

	if(device_list==NULL){
		output_log("ouch!!! no device is open\n");
		return B_ERROR;
	}

	acquire_sem(dev_list_lock);	
	*cookie=device_list;
	release_sem(dev_list_lock);	
	
	output_log("zen_device_open end ...\n");

	return B_OK;
}


/**
 *   my_device_read - handle read() calls
 *   
 */
static status_t
zen_device_read (void* cookie, off_t position, void *buf, size_t* num_bytes)
{
	mp3player_type *ptr=(mp3player_type*) cookie;
	char *heap= (char*) buf;
	if(device_list ==NULL)
		return -1;
	
	*num_bytes=ptr->size;	
	copy_to_userspace(heap,ptr);
	free_all_packets(&ptr->packet);
	ptr->packet=NULL;
	ptr->tail  =NULL;
	ptr->size=0;
	return 0;
}


/**
 *  we only use "zen_device_control" to read/write files
 *  because the mp3 player should be handles as database 
 *  hardware device and it's abnormal device!!
 */
static status_t
zen_device_write (void* cookie, off_t position, const void* buffer, size_t* num_bytes)
{
	*num_bytes = 0;				/* tell caller nothing was written */
	return B_IO_ERROR;
}


/***
 * zen_device_control - handle ioctl calls
 * only support most important functionalities
 * will be expanded later(25/03/2008)
 */
static status_t
zen_device_control (void* cookie, uint32 op, void* arg, size_t len)
{
	mp3player_type *dev= cookie;
	mp3player_type *ptr= dev;
	int i;
	char buff[1024];
	status_t  st;
	sprintf(buff,"zen_device_control control :%06x\n",op);
	output_log(buff);
	
	if(ptr != device_list || device_list == NULL){
		output_log("why ?! ... wrong pointer in control !! \n");
		return B_BAD_VALUE;
	}
	// if the status isn't those we have ...
	// assume the device status is overwritten by something else
	if(ptr->njb_status!=NJB_STATUS_HALT && ptr->njb_status!= NJB_STATUS_DOWNLOAD  && 
		ptr->njb_status!= NJB_STATUS_UPLOAD ){
		memset(buff,0,sizeof(buff));
		sprintf(buff,"njb_status %02x\n",ptr->njb_status);		
		output_log(buff);
		return B_BAD_VALUE;
	}

	switch(op){
		// here we download the tracks stored in Creative Zen Micro
		case NJB_CMD_GET_TRACK_TAG:
			output_log("now strip the tarck list");
			if(ptr->njb_status!=NJB_STATUS_HALT && ptr->size!=0){
				memset(buff,0,sizeof(buff));
				sprintf(buff,"njb_total_size %08x\n",ptr->size);		
				output_log(buff);
				return B_BAD_VALUE;
			}
			ptr->njb_status=NJB_STATUS_DOWNLOAD;
			ptr->num_packets=0;
			ptr->size=0;
			ptr->packet=NULL;
			
			if( send_njb_command(0, (void**) &ptr,(unsigned char*) njb3_get_track_tags,0x30) != B_OK ){
				output_log("error sending command... njb3_get_track_tags\n");
			}else{
				output_log("sending command... njb3_get_track_tags\n");
				recv( (void*) ptr,callback_burst_recv);
			}
			st= B_OK;
			break;
		case NJB_CMD_RECEIVE_FILE_BLOCK:
			// we also have to have a function to download
			// (we have to call *upload* due to the USB spec , I know this is very confusing )
			// each track
			output_log("now download the track \n");
			// we need 
			if(arg == NULL){
				return B_BAD_VALUE;
			}else if(ptr->njb_status!=NJB_STATUS_HALT && ptr->size!=0){
				memset(buff,0,sizeof(buff));
				sprintf(buff,"njb_total_size %08x\n",ptr->size);		
				output_log(buff);
				return B_BAD_VALUE;
			}else{
				mp3player_buffer *st_size=(mp3player_buffer*)arg;
				char com[0x11];
				memcpy(com,njb3_request_file,sizeof(com)-1);
				ptr->njb_status=NJB_STATUS_DOWNLOAD;
				ptr->num_packets=0;
				ptr->size=0;
				ptr->packet=NULL;
				// track id
				from_32bit_to_njb3_bytes(st_size->trackid,com+4);
				// offset
				from_32bit_to_njb3_bytes(st_size->offset,com+8);
				// chunk size 0x2000
				from_32bit_to_njb3_bytes(st_size->size, com + 12 );
				
				if( send_njb_command(0, (void**) &ptr,(unsigned char*) com,0x10) != B_OK ){
					output_log("error sending command... njb3_request_file \n");
				}else{
					recv( (void *) ptr , callback_normal_recv );
					recv( (void *) ptr , callback_burst_recv );
				}
			
				st= B_OK;
			}
			break;
		// creating file(track) tag is different from sending track
		// sending track
		case NJB_CMD_CREATE_FILE_TAG:
			{
				//
				mp3player_buffer *st_size=(mp3player_buffer*)arg;
				char *com=NULL; // command
				size_t sz=0x08 + st_size->size + 0x02;
				com = (char *) malloc(sz);
				memset(com,0,sz);
				output_log("now we are going to create file tag...\n");
				memcpy(com,njb3_create_file,0x08);
				from_16bit_to_njb3_bytes( st_size->datatype, com+0x06 );
				memcpy(com+0x08,st_size->buff,st_size->size);
				ptr->njb_status=NJB_STATUS_DOWNLOAD;
				ptr->num_packets=0;
				ptr->size=0;
				ptr->packet=NULL;
				
				if( send_njb_command(0, (void**) &ptr,(unsigned char*) com,sz) != B_OK ){
					output_log("error sending command... njb3_request_file \n");
					st=B_BAD_VALUE;
				}else{
					recv( (void *) ptr , callback_normal_recv );
					st=B_OK;
				}
				free(com);
			}
			
			
			break;
		case NJB_CMD_SEND_FILE_BLOCK_WITHOUT_HEADER:
			{
				mp3player_buffer *st_size=(mp3player_buffer*)arg;
				char *com=NULL;
				//char *data=NULL;
				output_log("ok ... sending file block\n");
				if(0<st_size->size){
					ptr->njb_status=NJB_STATUS_UPLOAD;
					output_log("ok ... we are going to send data only chunk\n");
					ptr->size_send = 0;
					if( (st=send_njb_command(0, (void**) &ptr,(unsigned char*) st_size->buff, st_size->size)) != B_OK){
						output_log("error sending command... njb3_request_file \n");
						st=B_BAD_VALUE;
					}
				}
			}
			break;
		case NJB_CMD_SEND_FILE_BLOCK_WITH_HEADER:
			{
				mp3player_buffer *st_size=(mp3player_buffer*)arg;
				char *com=NULL;
				//char *data=NULL;
				output_log("ok ... sending file block\n");
				//clear_pipe(ptr->conf->interface[0].active);
				
				if(0<st_size->size){
					
					sprintf(buff,"trackid %04x %04x %04x\n",st_size->trackid,st_size->filesize,ptr->size);
					output_log(buff);
					com = (char*) malloc(0x20);
					memset(com,0,0x20);
					memcpy(com,njb3_send_file_chunk,0x10);
					
					from_32bit_to_njb3_bytes(st_size->trackid, (unsigned char*) &com[4] );
					from_32bit_to_njb3_bytes(st_size->offset, (unsigned char*) &com[8] );
					
					sprintf(buff,"NJB_CMD_SEND_FILE_BLOCK_WITH_HEADER %06x %06x\n",st_size->filesize,st_size->size);
					output_log(buff);
					
					//if( 0x200000<st_size->filesize )
					//	from_32bit_to_njb3_bytes(0x200000, (unsigned char*) &com[12] );
					//else
					from_32bit_to_njb3_bytes(st_size->filesize, (unsigned char*) &com[12] );
					ptr->njb_status=NJB_STATUS_UPLOAD;
					output_log("ok ... we are going to send command header\n");
					ptr->size_send = 0;
					if( (st=send_njb_command(0, (void**) &ptr,(unsigned char*) com,0x10)) != B_OK ){
						output_log("error sending command... njb3_request_file \n");
						st=B_BAD_VALUE;
					}
					if( (st=send_njb_command(0, (void**) &ptr,(unsigned char*) st_size->buff, st_size->size)) != B_OK){
						output_log("error sending command... njb3_request_file \n");
						st=B_BAD_VALUE;
					}
					if(st==B_OK){
						//recv( (void *) ptr , callback_normal_recv );
						st=B_OK;
					}
					//st=B_OK;
				}else{
					sprintf(buff,"sending finished ...%04x \n",st_size->trackid);
					output_log(buff);
					com = (char*) malloc(0x08);
					memcpy(com,njb3_send_file_complete,0x08);
					from_32bit_to_njb3_bytes(st_size->trackid, (unsigned char*) (com + 4) );
					for(i=0;i<4;++i)
						sprintf(buff+i*3,"%02x ",com[4+i]);
					strcat(buff,"\n");
					output_log(buff);
					
					if( send_njb_command(0, (void**) &ptr,(unsigned char*) com,0x08) != B_OK){
						st=B_BAD_VALUE;
					}else if(recv( (void *) ptr , callback_normal_recv ) != B_OK){
						st=B_BAD_VALUE;
					}else
						st=B_OK;
				}
				if(com !=NULL)
					free(com);
			}
			break;
		// important !! 
		// This function should be always called 
		// when you are sending/receiving packets to/from
		// Creative Devices
		case NJB_CMD_GET_PACKET_SIZE_STATUS:
			{
				mp3player_buffer *st_size=(mp3player_buffer*)arg;
				output_log("NJB_CMD_GET_PACKET_SIZE_STATUS ...\n");
				acquire_sem(ptr->lock);
				st_size->offset=ptr->size_send;
				if(ptr->njb_status != NJB_STATUS_HALT ){
					st_size->status=ptr->njb_status;
					st_size->size=ptr->size;
					st=B_OK;
				}
				else if(ptr->size==0){
					st_size->size=0;
					st_size->status=NJB_STATUS_HALT;
					st=B_BAD_VALUE;
				}else{
					output_log("ok ... so the status is ... ?!\n");	
					st_size->status=ptr->njb_status;
					st_size->size  =ptr->size;
				
					st=B_OK;
				}
				release_sem(ptr->lock);
			}
			break;
		case NJB_CMD_DELETE_TRACK:
			{
				char *com = NULL;
				mp3player_buffer *st_size=(mp3player_buffer*)arg;
				com = (char *) malloc(0x08);
				memcpy(com,njb3_delitem,0x08);
				from_32bit_to_njb3_bytes(st_size->trackid, &com[4]);
				send_njb_command(0, (void**) &ptr,(unsigned char*) com,0x08);
				free(com);
				st=B_OK;
			}
			break;
		case NJB_CMD_STATUS_SEND :
			{
				mp3player_buffer *st_size=(mp3player_buffer*)arg;
				output_log("now we are going to check if the packets were sent to the device ...\n");
				sprintf(buff,"ptr->send_size:%06x\n",ptr->size_send); // because maximum of 0x200000
				output_log(buff);
				st_size->size = ptr->size_send;
			}
			break;
		case NJB_CMD_PING:
			send_njb_command(0,(void**) &ptr,  njb3_ping1 , 0x0c);
			recv( (void *) ptr , callback_normal_recv );
			st = B_OK;
			break;
		case NJB_INITIALISE : 
			send_njb_command(0,(void**) &ptr, (unsigned char *) njb3_read_keys , 0x13);
			recv( (void *) ptr , callback_normal_recv );
			st=B_OK;
			break;
		// we just check the return value from the device
		case NJB_CMD_CHECKN_SEND_FILE_BLOCK_STATUS:
			recv( (void *) ptr , callback_normal_recv );
			st=B_OK;
			break;
		default:
			st=B_BAD_VALUE;
			break;
	}

	return st;
}


/**
 *	zen_device_close - handle close() calls
 *  it's not that important
 */
static status_t
zen_device_close (void* cookie)
{
	mp3player_type *dev=cookie;
	mp3player_type *ptr=device_list;
	mp3player_type *ptr_del;
	output_log("zen_device_close is called\n");
	
	if(device_list == NULL || dev == NULL){
		output_log("ouch !! device_list is null ...zen_device_close \n");
		return B_ERROR;
	}

	if(dev==device_list ){

		acquire_sem(dev_list_lock);
		send_usb_command(ptr->dev,NJB3_CMD_RELEASE);
		if(dev->packet != NULL)
			free_all_packets(&dev->packet);	
		device_list=(void*) dev->next;
		delete_sem(ptr->lock);
		delete_sem(ptr->lock_send);
		release_sem(dev_list_lock);
		// delete all packets here!!
		clear_pipe(ptr->conf->interface[0].active);
		//delete_sem(dev_list_lock);
		free(dev);
		usb->uninstall_notify(drivername);
		put_module(B_USB_MODULE_NAME);
		output_log("zen_device_close dev==device_list\n");	
	}

	return B_OK;
}


/**
 *  my_device_free - called after the last device is closed, and after
 *  all i/o is complete.
 */
static status_t
zen_device_free (void* cookie)
{
	output_log("zen_device_free is called\n");
	//free(device_list);
	return B_OK;
}


/* -----
	function pointers for the device hooks entry points
----- */

device_hooks my_device_hooks = {
	zen_device_open, 			/* -> open entry point */
	zen_device_close, 			/* -> close entry point */
	zen_device_free,			/* -> free cookie */
	zen_device_control, 		/* -> control entry point */
	zen_device_read,			/* -> read entry point */
	zen_device_write			/* -> write entry point */
};

/**
 *  publish_devices - return a null-terminated array of devices
 *  supported by this driver.
 */
_EXPORT const char**
publish_devices()
{
	//kernel_debugger("publish devices");
	output_log("publish_devices is called \n");
	return(device_list == NULL) ? NULL : my_device_name;
}

/***
 *  find_device - return ptr to device hooks structure for a
 *  given device name
 */
_EXPORT device_hooks*
find_device(const char* name)
{
	char buff[1024];
	sprintf(buff,"find_device ... looking for %s\n",name);
	output_log(buff);
	if(device_list == NULL)
		return NULL;

	return &my_device_hooks ;
}

/**
 * only used for capture and release
 */
status_t send_usb_command(const usb_device *dev,uint8 com){
	size_t actual_len;
	status_t st;
	actual_len=0;
	st =usb->send_request(dev,USB_REQTYPE_CLASS,com
					  ,0,0,0,NULL,0,&actual_len);
	return ( st==B_OK)?B_OK : B_ERROR;
}

/**
 *  device_added is callback for USB Module
 *  this routine is fired when the device is connected
 */
static status_t 
device_added(const usb_device *dev, void **cookie)
{
	const usb_configuration_info *conf;
	usb_device_descriptor *dd = NULL;
	char tmp[8];
	char buff[1024];
	int i=0;
	status_t st;
	mp3player_type *ptr_dev = NULL;
	usb_interface_info *ii = NULL;
	usb_device0 *pDev = NULL;

	memset(buff,0,sizeof(buff));
	output_log("beginning of device_added\n");
	//
	//print_error(buff,0xFF);
	//kernel_debugger(buff);

	st=B_ERROR;
	// if device == NULL ... I doubt this  but often seen in other source
	if(dev == NULL )  return B_ERROR;

	// memory allocation for device node 
	if( !(ptr_dev= (mp3player_type *) malloc( sizeof(mp3player_type))) )
		return B_ERROR;
	
	memset(ptr_dev,0,sizeof(mp3player_type));
	ptr_dev->next = NULL;
	
	dd  = (usb_device_descriptor *) usb->get_device_descriptor(dev);
	if(dd == NULL)
		return B_ERROR;
	// caution , we need to set config again
	// after fetching default configutatin and tell USB Bus Manager
	// to setup... I dont know why
	if( (conf = usb->get_nth_configuration(dev,0)) != NULL) {
		ii=conf->interface[0].active;
		if(ii == NULL)
			return B_ERROR;

		clear_pipe(ii);
					
		output_log("found configuration \n");				
		if( (st=usb->set_configuration(dev,conf)) != B_OK){	
			output_log("config error\n");
			return B_ERROR;
		}

		

		
		sprintf(buff,"mp3player_sem%d",iCounter_Device++);
		acquire_sem(dev_list_lock);
		ptr_dev->lock=create_sem(1,buff);
		sprintf(buff,"mp3player_sem_send%d",iCounter_Device++);
		ptr_dev->lock_send=create_sem(1,buff);
		ptr_dev->conf=conf;
		ptr_dev->dev  = dev;
		ptr_dev->next = device_list;
		device_list = ptr_dev;
		*cookie=(void*) ptr_dev;
		release_sem(dev_list_lock);
		
		
		if(send_usb_command(dev,NJB3_CMD_CAPTURE) != B_OK)
			return B_ERROR;
		
		ptr_dev->njb_status=NJB_STATUS_HALT;							
		//  Capture NJB success !!
		ptr_dev->size=0;
		output_log("end of device_added\n");
		st = B_OK;
	}
	return st;
}

static status_t 
device_removed(void *cookie)
{
/*
	mp3player_type *dev=cookie;
	mp3player_type *ptr=device_list;
	mp3player_type *ptr_del;
	output_log("zen_device_close is called\n");
	
	if(device_list == NULL || dev == NULL){
		output_log("ouch !! device_list is null ...zen_device_close \n");
		return B_ERROR;
	}

	if(dev==device_list  ){
		acquire_sem(dev_list_lock);
		send_usb_command(ptr->dev,NJB3_CMD_RELEASE);
		device_list=(void*) dev->next;
		delete_sem(ptr->lock);
		delete_sem(ptr->lock_send);
		release_sem(dev_list_lock);
		if(dev->packet != NULL)
			free_all_packets(&dev->packet);
		// delete all packets here!!
		clear_pipe(ptr->conf->interface[0].active);
		delete_sem(dev_list_lock);
		free(dev);
		output_log("zen_device_close dev==device_list\n");
		return B_OK;		
	}
	*/
		
	output_log("device removed is called\n");
	return B_OK;
}